@@ -95,7 +95,7 @@ pub struct GraphicsOutput<'boot> {
 impl<'boot> GraphicsOutput<'boot> {
     /// Returns information for an available graphics mode that the graphics
     /// device and the set of active video output devices supports.
-    fn query_mode(&self, index: u32) -> Result<Mode> {
+    pub fn query_mode(&self, index: u32) -> Result<Mode> {
         let mut info_sz = 0;
         let mut info = ptr::null();
 
@@ -275,7 +275,7 @@ impl<'boot> GraphicsOutput<'boot> {
     fn check_blt_buffer_region(&self, region: BltRegion, dims: (usize, usize), buf_length: usize) {
         match region {
             BltRegion::Full => assert!(
-                dims.0.saturating_add(dims.1.saturating_mul(dims.0)) <= buf_length,
+                dims.1.saturating_mul(dims.0) <= buf_length,
                 "BltBuffer access out of bounds"
             ),
             BltRegion::SubRectangle {
