@@ -111,6 +111,13 @@ impl SystemTable<Boot> {
     /// `SystemTable<Boot>` view of the System Table and returning a more
     /// restricted `SystemTable<Runtime>` view as an output.
     ///
+    /// Once boot services are exited, the logger and allocator provided by
+    /// this crate can no longer be used. The logger should be disabled using
+    /// the [`Logger::disable`] method, and the allocator should be disabled by
+    /// calling [`alloc::exit_boot_services`]. Note that if the logger and
+    /// allocator were initialized with [`uefi_services::init`], they will be
+    /// disabled automatically when `exit_boot_services` is called.
+    ///
     /// The handle passed must be the one of the currently executing image,
     /// which is received by the entry point of the UEFI application. In
     /// addition, the application must provide storage for a memory map, which
@@ -132,6 +139,10 @@ impl SystemTable<Boot> {
     /// system table which more accurately reflects the state of the UEFI
     /// firmware following exit from boot services, along with a high-level
     /// iterator to the UEFI memory map.
+    ///
+    /// [`alloc::exit_boot_services`]: crate::alloc::exit_boot_services
+    /// [`Logger::disable`]: crate::logger::Logger::disable
+    /// [`uefi_services::init`]: https://docs.rs/uefi-services/latest/uefi_services/fn.init.html
     pub fn exit_boot_services(
         self,
         image: Handle,
@@ -208,6 +219,46 @@ impl SystemTable<Runtime> {
     pub unsafe fn runtime_services(&self) -> &RuntimeServices {
         self.table.runtime
     }
+
+    /// Changes the runtime addressing mode of EFI firmware from physical to virtual.
+    /// It is up to the caller to translate the old SystemTable address to a new virtual
+    /// address and provide it for this function.
+    /// See [`get_current_system_table_addr`]
+    ///
+    /// # Safety
+    ///
+    /// Setting new virtual memory map is unsafe and may cause undefined behaviors.
+    ///
+    /// [`get_current_system_table_addr`]: SystemTable::get_current_system_table_addr
+    pub unsafe fn set_virtual_address_map(
+        self,
+        map: &mut [MemoryDescriptor],
+        new_system_table_virtual_addr: u64,
+    ) -> Result<Self> {
+        // Unsafe Code Guidelines guarantees that there is no padding in an array or a slice
+        // between its elements if the element type is `repr(C)`, which is our case.
+        //
+        // See https://rust-lang.github.io/unsafe-code-guidelines/layout/arrays-and-slices.html
+        let map_size = core::mem::size_of_val(map);
+        let entry_size = core::mem::size_of::<MemoryDescriptor>();
+        let entry_version = crate::table::boot::MEMORY_DESCRIPTOR_VERSION;
+        let map_ptr = map.as_mut_ptr();
+        (self.table.runtime.set_virtual_address_map)(map_size, entry_size, entry_version, map_ptr)
+            .into_with_val(|| {
+                let new_table_ref =
+                    &mut *(new_system_table_virtual_addr as usize as *mut SystemTableImpl);
+                Self {
+                    table: new_table_ref,
+                    _marker: PhantomData,
+                }
+            })
+    }
+
+    /// Return the address of the SystemTable that resides in a UEFI runtime services
+    /// memory region.
+    pub fn get_current_system_table_addr(&self) -> u64 {
+        self.table as *const _ as usize as u64
+    }
 }
 
 /// The actual UEFI system table
